%
% Sequential quadratic Hamiltonian (SQH) scheme
%
% Main file for solving an elliptic linear optimal control problem with state constraints 
% formulated in a penalised cost functionalon a square [0,a]x[0,b] 
% see the paper
%
% A robust SQH scheme to solve non-smooth PDE optimal control problems
% by Tim Breitenbach and Alfio Borzi'
%
% Section 4.5 
%
% Problem specific parameters: 
% (a,b) interval length of the domain, 
% u_lo and u_up lower and upper bounds for the control, 
% alpha (weight quadratic cost term), 
% beta and s (weight discontinuous cost term), 
% c the upper bound for the state, gamma weight for the penalization term,
% N+1 number of intervals for the space discretisation, 
% kmax maximum number of iterations 

function [] = SQH_P6( )
OCP=struct('N',99,'a',0,'b',1,'u_lo',-100,'u_up',100,'alpha',0,'beta',10^-3,'s',20,'c',0.6,'gamma',100000);

%Tolerance for convergence
kappa=10^-10;                
%Algorithm paramters
eta=10^-9;
zeta=0.15;
sigma=50;
kmax=100000;        %Maximum iteration number
epsilon=(1/150);    %Initial guess for epsilon

h=(OCP.b-OCP.a)/(OCP.N+1);
[X1,X2]=meshgrid(OCP.a:h:OCP.b,OCP.a:h:OCP.b);
[X1u,X2u]=meshgrid(OCP.a+h:h:OCP.b-h,OCP.a+h:h:OCP.b-h);

%Assemble the discretised Laplacian
n=OCP.N+1;              
N=(n+1)*(n+1)-4*n;
v=ones(N,1);
v_block=[ones(n-2,1);0];
v_block=kron(ones(N/(n-1),1),v_block); 
A_hilf=spdiags([(-1/h^2)*v_block (-1/h^2)*[v_block(N); v_block(1:N-1)]],[ -1  1 ],N,N);
A=spdiags([(-1/h^2)*v  (4/h^2)*v  (-1/h^2)*v],[-(n-1)  0  n-1],N,N);
A=A+A_hilf;

yd=sin(2*pi*X1).*cos(2*pi*X2); %Desired state

u=zeros(OCP.N,OCP.N);          %Initial guess for the control 
u_old=u;
y=forward_y(u,A,OCP);          %Calculate the corresponding state variable y
y_old=y;
p=backward(y,yd,A,OCP);        %Calculate the corresponding adjoint state variable

J(1)=get_Jy(u,y,yd,OCP);       %Calculate the corresponding cost functional value

count_updates=1;               %Counter for the updates; counter_updates -1 for the number of updates
time_calc=cputime;             %Start time measurement

for k=1:kmax
     %Loop over all grid points
    for i=1:OCP.N    
       for j=1:OCP.N
            %Pointwise minimisation of the corresponding augmented
            %Hamiltonian with an exact formula
            pint=p(i+1,j+1);
            %Candidates for a minimum of the augmented Hamiltonian
            u1=min(max(OCP.s,(2*epsilon*u(i,j)-pint-OCP.beta)/(OCP.alpha+2*epsilon)),OCP.u_up); 
            u2=min(max(OCP.u_lo,(2*epsilon*u(i,j)-pint+OCP.beta)/(OCP.alpha+2*epsilon)),-OCP.s);
            u3=min(max(-OCP.s,(2*epsilon*u(i,j)-pint)/(OCP.alpha+2*epsilon)),OCP.s);            
            %Corresponding values of the augmented Hamiltonian
            H1=(OCP.alpha/2)*u1^2+OCP.beta*abs(u1)*(abs(u1)>OCP.s)+u1*pint+epsilon*((u1-u(i,j))^2);
            H2=(OCP.alpha/2)*u2^2+OCP.beta*abs(u2)*(abs(u2)>OCP.s)+u2*pint+epsilon*((u2-u(i,j))^2);
            H3=(OCP.alpha/2)*u3^2+u3*pint+epsilon*((u3-u(i,j))^2);
            %Choose the update for the control at the corresponding grid point that has the lowest 
            %value of the augmented Hamiltonian
            [~,pos]=min([H1,H2,H3]);        
            switch pos
                case 1
                    u(i,j)=u1;
                case 2
                    u(i,j)=u2;
                case 3
                    u(i,j)=u3;
            end
        end
    end
    y=forward_y(u,A,OCP);           %Calculation of the state equation with the updated control u
    J_int=get_Jy(u,y,yd,OCP);       %Corresponding cost functional value
    du=sum(sum((u_old-u).^2))*h*h;  %Evalute the norm square of the update
    if (J_int-J(count_updates)>-eta*du) 
        % If the improvemnt to the cost functional is not sufficiently large, 
        % take the old values of y and u, increase epsilon
        y=y_old;
        u=u_old;
        epsilon=epsilon*sigma;
    else                
        % If the improvment to the cost functional is sufficently large, 
        % take the new values of y, u, J, calculate p and decrease epsilon
        epsilon=epsilon*zeta;
        count_updates=count_updates+1;
        p=backward(y,yd,A,OCP);
        J(count_updates)=J_int;
        u_old=u;
        y_old=y;
        % Print current values
        fprintf('N.Iter. %i ; Value of J %e ; epsilon %e; ||u^(k+1)-u^k|| %e\n',count_updates-1,J_int,epsilon,du) 
    end
        
    if(du<kappa)    
        % If the norm square of the update is too small, then return the latest 
        % variable u that caused sufficient decrease of the cost functional 
        fprintf('Convergence achieved\n');
        u=u_old;
        break;
    end
    
end
ct=cputime-time_calc; %Stop time measurment for the calculation
fprintf('CPU-time in seconds needed: %f\n',ct)

%Plot y and u
figure
mesh(X1,X2,y);
  title('The state variable');
  xlabel('$x_1$','interpreter','latex');
  ylabel('$x_2$','interpreter','latex');
figure
mesh(X1u,X2u,u);
  title('The control variable');
  xlabel('$x_1$','interpreter','latex');
  ylabel('$x_2$','interpreter','latex');
figure
contour(X1u,X2u,u);
  title('Contour of the control variable');
  xlabel('$x_1$','interpreter','latex');
  ylabel('$x_2$','interpreter','latex');
  
%Plot convergence history
figure 
plot(J)
title('Convergence history of J');

fprintf('\n')
fprintf('Check numerical optimality of the SQH solution\n')
% Check solution for optimality; 
numeric_optimality(u,yd,OCP)

fprintf('Maximum value of y: %f\n', max(max(y)) )
end

