%
% Sequential quadratic Hamiltonian (SQH) scheme
%
% Main file for solving an elliptic linear optimal control problem with L^1 tracking term for the state 
% and max(y,0) in the state equationon on a square [0,a]x[0,b] 
% see the paper
%
% A robust SQH scheme to solve non-smooth PDE optimal control problems
% by Tim Breitenbach and Alfio Borzi'
%
% Section 4.4 
%
% Problem specific parameters: 
% (a,b) interval length of the domain, 
% u_lo and u_up lower and upper bounds for the control, 
% alpha (weight quadratic cost term), 
% beta weight for the non-convex cost term, 
% N+1 number of intervals for the space discretisation, 
% kmax maximum number of iterations 


function [ ] = SQH_P7( )
OCP=struct('N',99,'a',0,'b',1,'u_lo',-100,'u_up',100,'alpha',0,'beta',9*10^-2,'s',1);

%Tolerance for convergence
kappa=10^-8;                
%Algorithm paramters
eta=10^-9;
zeta=0.15;
sigma=50;
kmax=1000000;           %Maximum iteration number
epsilon=(1/150);        %Initial guess for epsilon

h=1/(OCP.N+1); %Step size for the discretisation
[X1,X2]=meshgrid(OCP.a:h:OCP.b,OCP.a:h:OCP.b); %Meshgrid for the plots
[X1u,X2u]=meshgrid(OCP.a+h:h:OCP.b-h,OCP.a+h:h:OCP.b-h); %Meshgrid for the plots of the control (only interior points)

yd=sin(2*pi*X1).*sin(2*pi*X2)+0.8; %Desired state
figure
mesh(X1,X2,yd)
u=zeros(OCP.N,OCP.N);  %Initial guess for the control 
y=forward_y(u,OCP);    %Calculate the corresponding state variable y
p=backward(y,yd,OCP);  %Calculate the corresponding adjoint state variable
u_old=u;
y_old=y;

J(1)=get_Jy(u,y,yd,OCP);%Calculate the corresponding cost functional value

count_updates=1;        %Counter for the updates; counter_updates -1 for the number of updates
time_calc=cputime;      %Start time measurement

for k=1:kmax
    %Loop over all grid points
    for i=1:OCP.N  
       for j=1:OCP.N
            %Pointwise minimisation of the corresponding augmented Hamiltonian
            u(i,j)=argMinH(u(i,j),p(i+1,j+1),epsilon,OCP); 
        end
    end
    y=forward_y(u,OCP);             %Calculation of the state equation with the updated control u
    J_int=get_Jy(u,y,yd,OCP);       %Corresponding cost functional value
    du=sum(sum((u_old-u).^2))*h*h;  %Evalute the norm square of the update
    if (J_int-J(count_updates)>-eta*du) 
        % If the improvemnt to the cost functional is not sufficiently large, 
        % take the old values of y and u, increase epsilon
        y=y_old;
        u=u_old;
        epsilon=epsilon*sigma;
    else   
        % If the improvment to the cost functional is sufficently large, 
        % take the new values of y, u, J, calculate p and decrease epsilon
        epsilon=epsilon*zeta;
        count_updates=count_updates+1;
        p=backward(y,yd,OCP);
        J(count_updates)=J_int;
        u_old=u;
        y_old=y; 
        % Print current values
        fprintf('N.Iter. %i ; Value of J %e ; epsilon %e; ||u^(k+1)-u^k|| %e\n',count_updates-1,J_int,epsilon,du)
    end
        
    if(du<kappa)   
        % If the norm square of the update is too small, then return the latest 
        % variable u that caused sufficient decrease of the cost functional 
        fprintf('Convergence achieved\n');
        u=u_old;
        break;
    end    
end
ct=cputime-time_calc;            %Stop time measurment for the calculation 
fprintf('CPU-time in seconds needed: %f\n',ct)

%Plot y and u
figure
mesh(X1,X2,y);
  %title('The state variable');
  xlabel('$x_1$','interpreter','latex');
  ylabel('$x_2$','interpreter','latex');
figure
mesh(X1u,X2u,u);
  %title('The control variable');
  xlabel('$x_1$','interpreter','latex');
  ylabel('$x_2$','interpreter','latex');
figure
contour(X1u,X2u,u);
  %title('Contour of the control variable');
  xlabel('$x_1$','interpreter','latex');
  ylabel('$x_2$','interpreter','latex');
  
%Plot convergence history
figure 
plot(J)
title('Convergence history of J');

fprintf('\n')
fprintf('Check numerical optimality of the SQH solution\n')
% Check solution for optimality; 
numeric_optimality(u,yd,OCP)
end